from collections import defaultdict
import json
from typing import Dict, List, Optional


class JsonSerializable(object):
    def to_json(self):
        return self.payload

    def __str__(self):
        return json.dumps(self.payload, indent=4)


class _LifecycleHook(JsonSerializable):
    # https://argoproj.github.io/argo-workflows/fields/#lifecyclehook

    def __init__(self, name):
        tree = lambda: defaultdict(tree)
        self.name = name
        self.payload = tree()

    def expression(self, expression):
        self.payload["expression"] = str(expression)
        return self

    def template(self, template):
        self.payload["template"] = template
        return self


class _Template(JsonSerializable):
    # https://argoproj.github.io/argo-workflows/fields/#template

    def __init__(self, name):
        tree = lambda: defaultdict(tree)
        self.name = name
        self.payload = tree()
        self.payload["name"] = name

    def http(self, http):
        self.payload["http"] = http.to_json()
        return self

    def script(self, script):
        self.payload["script"] = script.to_json()
        return self

    def container(self, container):
        self.payload["container"] = container
        return self

    def service_account_name(self, service_account_name):
        self.payload["serviceAccountName"] = service_account_name
        return self


class Hook(object):
    """
    Abstraction for Argo Workflows exit hooks.
    A hook consists of a Template, and one or more LifecycleHooks that trigger the template
    """

    template: "_Template"
    lifecycle_hooks: List["_LifecycleHook"]


class _HttpSpec(JsonSerializable):
    # https://argoproj.github.io/argo-workflows/fields/#http

    def __init__(self, method):
        tree = lambda: defaultdict(tree)
        self.payload = tree()
        self.payload["method"] = method
        self.payload["headers"] = []

    def header(self, header, value):
        self.payload["headers"].append({"name": header, "value": value})
        return self

    def body(self, body):
        self.payload["body"] = str(body)
        return self

    def url(self, url):
        self.payload["url"] = url
        return self

    def success_condition(self, success_condition):
        self.payload["successCondition"] = success_condition
        return self


# HTTP hook
class HttpExitHook(Hook):
    def __init__(
        self,
        name: str,
        url: str,
        method: str = "GET",
        headers: Optional[Dict] = None,
        body: Optional[str] = None,
        on_success: bool = False,
        on_error: bool = False,
    ):
        self.template = _Template(name)
        http = _HttpSpec(method).url(url)
        if headers is not None:
            for header, value in headers.items():
                http.header(header, value)

        if body is not None:
            http.body(body)

        self.template.http(http)

        self.lifecycle_hooks = []

        if on_success and on_error:
            raise Exception("Set only one of the on_success/on_error at a time.")

        if on_success:
            self.lifecycle_hooks.append(
                _LifecycleHook(name)
                .expression("workflow.status == 'Succeeded'")
                .template(self.template.name)
            )

        if on_error:
            self.lifecycle_hooks.append(
                _LifecycleHook(name)
                .expression("workflow.status == 'Error' || workflow.status == 'Failed'")
                .template(self.template.name)
            )

        if not on_success and not on_error:
            # add an expressionless lifecycle hook
            self.lifecycle_hooks.append(_LifecycleHook(name).template(name))


class ExitHookHack(Hook):
    # Warning: terrible hack to workaround a bug in Argo Workflow where the
    #          templates listed above do not execute unless there is an
    #          explicit exit hook. as and when this bug is patched, we should
    #          remove this effectively no-op template.
    # Note: We use the Http template because changing this to an actual no-op container had the side-effect of
    # leaving LifecycleHooks in a pending state even when they have finished execution.
    def __init__(
        self,
        url,
        headers=None,
        body=None,
    ):
        self.template = _Template("exit-hook-hack")
        http = _HttpSpec("GET").url(url)
        if headers is not None:
            for header, value in headers.items():
                http.header(header, value)

        if body is not None:
            http.body(json.dumps(body))

        http.success_condition("true == true")

        self.template.http(http)

        self.lifecycle_hooks = []

        # add an expressionless lifecycle hook
        self.lifecycle_hooks.append(_LifecycleHook("exit").template("exit-hook-hack"))


class ContainerHook(Hook):
    def __init__(
        self,
        name: str,
        container: Dict,
        service_account_name: str = None,
        on_success: bool = False,
        on_error: bool = False,
    ):
        self.template = _Template(name)

        if service_account_name is not None:
            self.template.service_account_name(service_account_name)

        self.template.container(container)

        self.lifecycle_hooks = []

        if on_success and on_error:
            raise Exception("Set only one of the on_success/on_error at a time.")

        if on_success:
            self.lifecycle_hooks.append(
                _LifecycleHook(name)
                .expression("workflow.status == 'Succeeded'")
                .template(self.template.name)
            )

        if on_error:
            self.lifecycle_hooks.append(
                _LifecycleHook(name)
                .expression("workflow.status == 'Error' || workflow.status == 'Failed'")
                .template(self.template.name)
            )

        if not on_success and not on_error:
            # add an expressionless lifecycle hook
            self.lifecycle_hooks.append(_LifecycleHook(name).template(name))



import time

from contextlib import contextmanager

from .metaflow_config import PROFILE_FROM_START

init_time = None


if PROFILE_FROM_START:

    def from_start(msg: str):
        global init_time
        if init_time is None:
            init_time = time.time()
        print("From start: %s took %dms" % (msg, int((time.time() - init_time) * 1000)))

else:

    def from_start(_msg: str):
        pass


@contextmanager
def profile(label, stats_dict=None):
    if stats_dict is None:
        print("PROFILE: %s starting" % label)
    start = time.time()
    yield
    took = int((time.time() - start) * 1000)
    if stats_dict is None:
        print("PROFILE: %s completed in %dms" % (label, took))
    else:
        stats_dict[label] = stats_dict.get(label, 0) + took



import json
import os
import re
import tempfile
from typing import Tuple, Dict

from metaflow.decorators import StepDecorator
from metaflow.metadata_provider import MetaDatum
from metaflow.metaflow_current import current
from metaflow.user_configs.config_options import ConfigInput
from metaflow.user_configs.config_parameters import dump_config_values
from metaflow.util import to_unicode

from .component_serializer import CardComponentCollector, get_card_class
from .card_creator import CardCreator
from .exception import CARD_ID_PATTERN, TYPE_CHECK_REGEX

ASYNC_TIMEOUT = 30


def warning_message(message, logger=None, ts=False):
    msg = "[@card WARNING] %s" % message
    if logger:
        logger(msg, timestamp=ts, bad=True)


class MetadataStateManager(object):
    def __init__(self, info_func):
        self._info_func = info_func
        self._metadata_registered = {}

    def register_metadata(self, card_uuid) -> Tuple[bool, Dict]:
        info = self._info_func()
        # Check that metadata was not written yet. We only want to write once.
        if (
            info is None
            or info.get(card_uuid) is None
            or self._metadata_registered.get(card_uuid)
        ):
            return False, {}
        self._metadata_registered[card_uuid] = True
        return True, info.get(card_uuid)


class CardDecorator(StepDecorator):
    """
    Creates a human-readable report, a Metaflow Card, after this step completes.

    Note that you may add multiple `@card` decorators in a step with different parameters.

    Parameters
    ----------
    type : str, default 'default'
        Card type.
    id : str, optional, default None
        If multiple cards are present, use this id to identify this card.
    options : Dict[str, Any], default {}
        Options passed to the card. The contents depend on the card type.
    timeout : int, default 45
        Interrupt reporting if it takes more than this many seconds.

    MF Add To Current
    -----------------
    card -> metaflow.plugins.cards.component_serializer.CardComponentCollector
        The `@card` decorator makes the cards available through the `current.card`
        object. If multiple `@card` decorators are present, you can add an `ID` to
        distinguish between them using `@card(id=ID)` as the decorator. You will then
        be able to access that specific card using `current.card[ID].

        Methods available are `append` and `extend`

        @@ Returns
        -------
        CardComponentCollector
            The or one of the cards attached to this step.
    """

    _GLOBAL_CARD_INFO = {}

    name = "card"
    defaults = {
        "type": "default",
        "options": {},
        "scope": "task",
        "rank": None,  # Can be one of "high", "medium", "low". Can help derive ordering on the UI.
        "timeout": 45,
        "id": None,
        "save_errors": True,
        "customize": False,
        "refresh_interval": 5,
    }
    allow_multiple = True

    total_decos_on_step = {}

    step_counter = 0

    _called_once = {}

    card_creator = None

    _config_values = None

    _config_file_name = None

    task_finished_decos = 0

    def __init__(self, *args, **kwargs):
        super(CardDecorator, self).__init__(*args, **kwargs)
        self._task_datastore = None
        self._environment = None
        self._metadata = None
        self._logger = None
        self._is_editable = False
        self._card_uuid = None
        self._user_set_card_id = None
        self._metadata_registered = False

    @classmethod
    def _set_card_creator(cls, card_creator):
        cls.card_creator = card_creator

    def _is_event_registered(self, evt_name):
        return evt_name in self._called_once

    @classmethod
    def _register_event(cls, evt_name):
        if evt_name not in cls._called_once:
            cls._called_once[evt_name] = True

    @classmethod
    def _set_card_counts_per_step(cls, step_name, total_count):
        cls.total_decos_on_step[step_name] = total_count

    @classmethod
    def _increment_step_counter(cls):
        cls.step_counter += 1

    @classmethod
    def _increment_completed_counter(cls):
        cls.task_finished_decos += 1

    @classmethod
    def _set_config_values(cls, config_values):
        cls._config_values = config_values

    @classmethod
    def _set_config_file_name(cls, flow):
        # Only create a config file from the very first card decorator.
        if cls._config_values and not cls._config_file_name:
            with tempfile.NamedTemporaryFile(
                mode="w", encoding="utf-8", delete=False
            ) as config_file:
                config_value = dump_config_values(flow)
                json.dump(config_value, config_file)
                cls._config_file_name = config_file.name

    @classmethod
    def _register_card_info(cls, **kwargs):
        if not kwargs.get("card_uuid"):
            raise ValueError("card_uuid is required")
        cls._GLOBAL_CARD_INFO[kwargs["card_uuid"]] = kwargs

    @classmethod
    def all_cards_info(cls):
        return cls._GLOBAL_CARD_INFO.copy()

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        self._flow_datastore = flow_datastore
        self._environment = environment
        self._logger = logger

        self.card_options = None

        # We check for configuration options. We do this here before they are
        # converted to properties.
        self._set_config_values(
            [
                (config.name, ConfigInput.make_key_name(config.name))
                for _, config in flow._get_parameters()
                if config.IS_CONFIG_PARAMETER
            ]
        )

        self.card_options = self.attributes["options"]

        evt_name = "step-init"
        # `'%s-%s'%(evt_name,step_name)` ensures that we capture this once per @card per @step.
        # Since there can be many steps checking if event is registered for `evt_name` will only make it check it once for all steps.
        # Hence, we have `_is_event_registered('%s-%s'%(evt_name,step_name))`
        self._is_runtime_card = False
        evt = "%s-%s" % (evt_name, step_name)
        if not self._is_event_registered(evt):
            # We set the total count of decorators so that we can use it for
            # when calling the finalize function of CardComponentCollector
            # We set the total @card per step via calling `_set_card_counts_per_step`.
            other_card_decorators = [
                deco for deco in decorators if isinstance(deco, self.__class__)
            ]
            self._set_card_counts_per_step(step_name, len(other_card_decorators))
            self._register_event(evt)

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        self._task_datastore = task_datastore
        self._metadata = metadata

        # If we have configs, we need to dump them to a file so we can re-use them
        # when calling the card creation subprocess.
        # Since a step can contain multiple card decorators, and all the card creation processes
        # will reference the same config file (because of how the CardCreator is created (only single class instance)),
        # we need to ensure that a single config file is being referenced for all card create commands.
        # This config file will be removed when the last card decorator has finished creating its card.
        self._set_config_file_name(flow)
        # The MetadataStateManager is used to track the state of the metadata registration.
        # It is there to ensure that we only register metadata for the card once. This is so that we
        # avoid any un-necessary metadata writes because the create command can be called multiple times during the
        # card creation process.
        self._metadata_state_manager = MetadataStateManager(self.all_cards_info)

        card_type = self.attributes["type"]
        card_class = get_card_class(card_type)

        self._is_runtime_card = False
        if card_class is not None:  # Card type was not found
            if card_class.ALLOW_USER_COMPONENTS:
                self._is_editable = True
            self._is_runtime_card = card_class.RUNTIME_UPDATABLE

        # We have a step counter to ensure that on calling the final card decorator's `task_pre_step`
        # we call a `finalize` function in the `CardComponentCollector`.
        # This can help ensure the behaviour of the `current.card` object is according to specification.
        self._increment_step_counter()
        self._user_set_card_id = self.attributes["id"]
        if (
            self.attributes["id"] is not None
            and re.match(CARD_ID_PATTERN, self.attributes["id"]) is None
        ):
            # There should be a warning issued to the user that `id` doesn't match regex pattern
            # Since it is doesn't match pattern, we need to ensure that `id` is not accepted by `current`
            # and warn users that they cannot use id for their arguments.
            wrn_msg = (
                "@card with id '%s' doesn't match REGEX pattern. "
                "Adding custom components to cards will not be accessible via `current.card['%s']`. "
                "Please create `id` of pattern %s. "
            ) % (self.attributes["id"], self.attributes["id"], TYPE_CHECK_REGEX)
            warning_message(wrn_msg, self._logger)
            self._user_set_card_id = None

        # As we have multiple decorators,
        # we need to ensure that `current.card` has `CardComponentCollector` instantiated only once.
        if not self._is_event_registered("pre-step"):
            self._register_event("pre-step")
            self._set_card_creator(
                CardCreator(
                    self._create_top_level_args(flow),
                    self._metadata_state_manager.register_metadata,
                )
            )

            current._update_env(
                {"card": CardComponentCollector(self._logger, self.card_creator)}
            )

        # this line happens because of decospecs parsing.
        customize = False
        if str(self.attributes["customize"]) == "True":
            customize = True

        card_metadata = current.card._add_card(
            self.attributes["type"],
            self._user_set_card_id,
            self.attributes,
            self.card_options,
            editable=self._is_editable,
            customize=customize,
            runtime_card=self._is_runtime_card,
            refresh_interval=self.attributes["refresh_interval"],
        )
        self._card_uuid = card_metadata["uuid"]

        self._register_card_info(
            card_uuid=self._card_uuid,
            rank=self.attributes["rank"],
            type=self.attributes["type"],
            options=self.card_options,
            is_editable=self._is_editable,
            is_runtime_card=self._is_runtime_card,
            refresh_interval=self.attributes["refresh_interval"],
            customize=customize,
            id=self._user_set_card_id,
        )

        # This means that we are calling `task_pre_step` on the last card decorator.
        # We can now `finalize` method in the CardComponentCollector object.
        # This will set up the `current.card` object for usage inside `@step` code.
        if self.step_counter == self.total_decos_on_step[step_name]:
            current.card._finalize()

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_user_code_retries
    ):
        create_options = dict(
            card_uuid=self._card_uuid,
            user_set_card_id=self._user_set_card_id,
            runtime_card=self._is_runtime_card,
            decorator_attributes=self.attributes,
            card_options=self.card_options,
            logger=self._logger,
        )
        if is_task_ok:
            self.card_creator.create(mode="render", final=True, **create_options)
            self.card_creator.create(mode="refresh", final=True, **create_options)

        self._cleanup(step_name)

    @staticmethod
    def _options(mapping):
        for k, v in mapping.items():
            if v:
                k = k.replace("_", "-")
                v = v if isinstance(v, (list, tuple, set)) else [v]
                for value in v:
                    yield "--%s" % k
                    if not isinstance(value, bool):
                        if isinstance(value, tuple):
                            for val in value:
                                yield to_unicode(val)
                        else:
                            yield to_unicode(value)

    def _create_top_level_args(self, flow):
        top_level_options = {
            "quiet": True,
            "metadata": self._metadata.TYPE,
            "environment": self._environment.TYPE,
            "datastore": self._flow_datastore.TYPE,
            "datastore-root": self._flow_datastore.datastore_root,
            "no-pylint": True,
            "event-logger": "nullSidecarLogger",
            "monitor": "nullSidecarMonitor",
            # We don't provide --with as all execution is taking place in
            # the context of the main process
        }
        if self._config_values:
            top_level_options["config-value"] = self._config_values
            top_level_options["local-config-file"] = self._config_file_name

        return list(self._options(top_level_options))

    def _cleanup(self, step_name):
        self._increment_completed_counter()
        if self.task_finished_decos == self.total_decos_on_step[step_name]:
            # Unlink the config file if it exists
            if self._config_file_name:
                try:
                    os.unlink(self._config_file_name)
                except Exception as e:
                    pass



import time

from metaflow.metaflow_config import SFN_DYNAMO_DB_TABLE


class DynamoDbClient(object):
    def __init__(self):
        from ..aws_client import get_aws_client

        self._client = get_aws_client("dynamodb")
        self.name = SFN_DYNAMO_DB_TABLE

    def save_foreach_cardinality(self, foreach_split_task_id, foreach_cardinality, ttl):
        return self._client.put_item(
            TableName=self.name,
            Item={
                "pathspec": {"S": foreach_split_task_id},
                "for_each_cardinality": {
                    "NS": list(map(str, range(foreach_cardinality)))
                },
                "ttl": {"N": str(ttl)},
            },
        )

    def save_parent_task_id_for_foreach_join(
        self, foreach_split_task_id, foreach_join_parent_task_id
    ):
        ex = None
        for attempt in range(10):
            try:
                return self._client.update_item(
                    TableName=self.name,
                    Key={"pathspec": {"S": foreach_split_task_id}},
                    UpdateExpression="ADD parent_task_ids_for_foreach_join :val",
                    ExpressionAttributeValues={
                        ":val": {"SS": [foreach_join_parent_task_id]}
                    },
                )
            except self._client.exceptions.ClientError as error:
                ex = error
                if (
                    error.response["Error"]["Code"]
                    == "ProvisionedThroughputExceededException"
                ):
                    # hopefully, enough time for AWS to scale up! otherwise
                    # ensure sufficient on-demand throughput for dynamo db
                    # is provisioned ahead of time
                    sleep_time = min((2**attempt) * 10, 60)
                    time.sleep(sleep_time)
                else:
                    raise
        raise ex

    def get_parent_task_ids_for_foreach_join(self, foreach_split_task_id):
        response = self._client.get_item(
            TableName=self.name,
            Key={"pathspec": {"S": foreach_split_task_id}},
            ProjectionExpression="parent_task_ids_for_foreach_join",
            ConsistentRead=True,
        )
        return response["Item"]["parent_task_ids_for_foreach_join"]["SS"]



from math import inf
import sys
from metaflow.util import decompress_list, compress_list
import base64


def generate_input_paths(input_paths, skippable_steps):
    # => run_id/step/:foo,bar
    # input_paths are base64 encoded due to Argo shenanigans
    decoded = base64.b64decode(input_paths).decode("utf-8")
    paths = decompress_list(decoded)

    # some of the paths are going to be malformed due to never having executed per conditional.
    # strip these out of the list.

    # all pathspecs of leading steps that executed.
    trimmed = [path for path in paths if not "{{" in path]

    # pathspecs of leading steps that are conditional, and should be used instead of non-conditional ones
    # e.g. the case of skipping switches: start -> case_step -> conditional_a or end
    conditionals = [
        path for path in trimmed if not any(step in path for step in skippable_steps)
    ]
    pathspecs_to_use = conditionals if conditionals else trimmed
    return compress_list(pathspecs_to_use, zlibmin=inf)


if __name__ == "__main__":
    input_paths = sys.argv[1]
    try:
        skippable_steps = sys.argv[2].split(",")
    except IndexError:
        skippable_steps = []

    print(generate_input_paths(input_paths, skippable_steps))



from metaflow.plugins.secrets import SecretsProvider
import re
import base64
import codecs
from urllib.parse import urlparse
from metaflow.exception import MetaflowException
import sys
from metaflow.metaflow_config import AZURE_KEY_VAULT_PREFIX
from metaflow.plugins.azure.azure_credential import (
    create_cacheable_azure_credential,
)


class MetaflowAzureKeyVaultBadVault(MetaflowException):
    """Raised when the secretid is fully qualified but does not have the right key vault domain"""


class MetaflowAzureKeyVaultBadSecretType(MetaflowException):
    """Raised when the secret type is anything except secrets"""


class MetaflowAzureKeyVaultBadSecretPath(MetaflowException):
    """Raised when the secret path does not match to expected length"""


class MetaflowAzureKeyVaultBadSecretName(MetaflowException):
    """Raised when the secret name does not match expected pattern"""


class MetaflowAzureKeyVaultBadSecretVersion(MetaflowException):
    """Raised when the secret version does not match expected pattern"""


class MetaflowAzureKeyVaultBadSecret(MetaflowException):
    """Raised when the secret does not match supported patterns in Metaflow"""


class AzureKeyVaultSecretsProvider(SecretsProvider):
    TYPE = "az-key-vault"
    key_vault_domains = [
        ".vault.azure.net",
        ".vault.azure.cn",
        ".vault.usgovcloudapi.net",
        ".vault.microsoftazure.de",
    ]
    supported_vault_object_types = ["secrets"]

    # https://learn.microsoft.com/en-us/azure/key-vault/general/about-keys-secrets-certificates has details on vault name structure
    # Vault name and Managed HSM pool name must be a 3-24 character string, containing only 0-9, a-z, A-Z, and not consecutive -.
    def _is_valid_vault_name(self, vault_name):
        vault_name_pattern = r"^(?!.*--)[a-zA-Z0-9-]{3,24}$"
        return re.match(vault_name_pattern, vault_name) is not None

    # The type of the object can be, "keys", "secrets", or "certificates".
    # Currently only secrets will be supported
    def _is_valid_object_type(self, secret_type):
        for type in self.supported_vault_object_types:
            if secret_type == type:
                return True
        return False

    # The secret name must be a 1-127 character string, starting with a letter and containing only 0-9, a-z, A-Z, and -.
    def _is_valid_secret_name(self, secret_name):
        secret_name_pattern = r"^[a-zA-Z][a-zA-Z0-9-]{0,126}$"
        return re.match(secret_name_pattern, secret_name) is not None

    # An object-version is a system-generated, 32 character string identifier that is optionally used to address a unique version of an object.
    def _is_valid_object_version(self, secret_version):
        object_version_pattern = r"^[a-zA-Z0-9]{32}$"
        return re.match(object_version_pattern, secret_version) is not None

    # This function will check if the secret_id is fully qualified url. It will return True iff the secret_id is of the form:
    # https://myvault.vault.azure.net/secrets/mysecret/ec96f02080254f109c51a1f14cdb1931 OR
    # https://myvault.vault.azure.net/secrets/mysecret/
    # validating the above as per recommendations in https://devblogs.microsoft.com/azure-sdk/guidance-for-applications-using-the-key-vault-libraries/
    def _is_secret_id_fully_qualified_url(self, secret_id):
        # if the secret_id is None/empty/does not start with https then return false
        if secret_id is None or secret_id == "" or not secret_id.startswith("https://"):
            return False
        try:
            parsed_vault_url = urlparse(secret_id)
        except ValueError:
            print("invalid vault url", file=sys.stderr)
            return False
        hostname = parsed_vault_url.netloc

        k_v_domain_found = False
        actual_k_v_domain = ""
        for k_v_domain in self.key_vault_domains:
            if k_v_domain in hostname:
                k_v_domain_found = True
                actual_k_v_domain = k_v_domain
                break
        if not k_v_domain_found:
            # the secret_id started with https:// however the key_vault_domains
            # were not present in the secret_id which means
            raise MetaflowAzureKeyVaultBadVault("bad key vault domain %s" % secret_id)

        # given the secret_id seems to have a valid key vault domain
        # lets verify that the vault name corresponds to its regex.
        vault_name = hostname[: -len(actual_k_v_domain)]
        # verify the vault name pattern
        if not self._is_valid_vault_name(vault_name):
            raise MetaflowAzureKeyVaultBadVault("bad key vault name %s" % vault_name)

        path_parts = parsed_vault_url.path.strip("/").split("/")
        total_path_parts = len(path_parts)
        if total_path_parts < 2 or total_path_parts > 3:
            raise MetaflowAzureKeyVaultBadSecretPath(
                "bad secret uri path %s" % path_parts
            )

        object_type = path_parts[0]
        if not self._is_valid_object_type(object_type):
            raise MetaflowAzureKeyVaultBadSecretType("bad secret type %s" % object_type)

        secret_name = path_parts[1]
        if not self._is_valid_secret_name(secret_name=secret_name):
            raise MetaflowAzureKeyVaultBadSecretName("bad secret name %s" % secret_name)

        if total_path_parts == 3:
            if not self._is_valid_object_version(path_parts[2]):
                raise MetaflowAzureKeyVaultBadSecretVersion(
                    "bad secret version %s" % path_parts[2]
                )

        return True

    # This function will validate the correctness of the partial secret id.
    # It will attempt to construct the fully qualified secret URL internally and
    # call the _is_secret_id_fully_qualified_url to check validity
    def _is_partial_secret_valid(self, secret_id):
        secret_parts = secret_id.strip("/").split("/")
        total_secret_parts = len(secret_parts)
        if total_secret_parts < 1 or total_secret_parts > 2:
            return False

        # since the secret_id is supposedly a partial id, the AZURE_KEY_VAULT_PREFIX
        # must be set.
        if not AZURE_KEY_VAULT_PREFIX:
            raise ValueError(
                "cannot use simple secret id without setting METAFLOW_AZURE_KEY_VAULT_PREFIX. %s"
                % AZURE_KEY_VAULT_PREFIX
            )
        domain = AZURE_KEY_VAULT_PREFIX.rstrip("/")
        full_secret = "%s/secrets/%s" % (domain, secret_id)
        if not self._is_secret_id_fully_qualified_url(full_secret):
            return False

        return True

    def _sanitize_key_as_env_var(self, key):
        """
        Sanitize a key as an environment variable name.
        This is purely a convenience trade-off to cover common cases well, vs. introducing
        ambiguities (e.g. did the final '_' come from '.', or '-' or is original?).

        1/27/2023(jackie):

        We start with few rules and should *sparingly* add more over time.
        Also, it's TBD whether all possible providers will share the same sanitization logic.
        Therefore we will keep this function private for now
        """
        return key.replace("-", "_").replace(".", "_").replace("/", "_")

    def get_secret_as_dict(self, secret_id, options={}, role=None):
        # https://learn.microsoft.com/en-us/azure/app-service/app-service-key-vault-references?tabs=azure-cli has a lot of details on
        # the patterns used in key vault
        # Vault names and Managed HSM pool names are selected by the user and are globally unique.
        # Vault name and Managed HSM pool name must be a 3-24 character string, containing only 0-9, a-z, A-Z, and not consecutive -.
        # object-type	The type of the object. As of 05/08/24 only "secrets", are supported
        # object-name	An object-name is a user provided name for and must be unique within a key vault. The name must be a 1-127 character string, starting with a letter and containing only 0-9, a-z, A-Z, and -.
        # object-version	An object-version is a system-generated, 32 character string identifier that is optionally used to address a unique version of an object.

        # We allow these forms of secret_id:
        #
        # 1. Full path like https://<key-vault-name><.vault-domain>/secrets/<secret-name>/<secret-version>. This is what you
        # see in Azure portal and is easy to copy paste.
        #
        # 2. Full path but without the version like https://<key-vault-name><.vault-domain>/secrets/<secret-name>
        #
        # 3. Simple string like mysecret. This corresponds to the SecretName.
        #
        # 4. Simple string with <secret-name>/<secret-version> suffix like mysecret/123

        # The latter two forms require METAFLOW_AZURE_KEY_VAULT_PREFIX to be set.

        # if the secret_id is None/empty/does not start with https then return false
        if secret_id is None or secret_id == "":
            raise MetaflowAzureKeyVaultBadSecret("empty secret id is not supported")

        # check if the passed in secret is a short-form ( #3/#4 in the above comment)
        if not secret_id.startswith("https://"):
            # check if the secret_id is of form `secret_name` OR `secret_name/secret_version`
            if not self._is_partial_secret_valid(secret_id=secret_id):
                raise MetaflowAzureKeyVaultBadSecret(
                    "unsupported partial secret %s" % secret_id
                )

            domain = AZURE_KEY_VAULT_PREFIX.rstrip("/")
            full_secret = "%s/secrets/%s" % (domain, secret_id)

        # if the secret id is passed as a URL - then check if the url is fully qualified
        if secret_id.startswith("https://"):
            if not self._is_secret_id_fully_qualified_url(secret_id=secret_id):
                raise MetaflowException("unsupported secret %s" % secret_id)
            full_secret = secret_id

        # at this point I know that the secret URL is good so we can start creating the Secret Client
        az_credentials = create_cacheable_azure_credential()
        res = urlparse(full_secret)
        az_vault_url = "%s://%s" % (
            res.scheme,
            res.netloc,
        )  # https://myvault.vault.azure.net
        secret_data = res.path.strip("/").split("/")[1:]
        secret_name = secret_data[0]
        secret_version = None
        if len(secret_data) > 1:
            secret_version = secret_data[1]

        from azure.keyvault.secrets import SecretClient

        client = SecretClient(vault_url=az_vault_url, credential=az_credentials)

        key_vault_secret_val = client.get_secret(
            name=secret_name, version=secret_version
        )

        result = {}

        if options.get("env_var_name") is not None:
            env_var_name = options["env_var_name"]
            sanitized_key = self._sanitize_key_as_env_var(env_var_name)
        else:
            sanitized_key = self._sanitize_key_as_env_var(key_vault_secret_val.name)

        response_payload = key_vault_secret_val.value
        result[sanitized_key] = response_payload
        return result



import inspect


def get_methods(class_object):
    all_attributes = {}
    all_methods = {}
    if isinstance(class_object, type):
        mros = list(reversed(type(class_object).__mro__)) + list(
            reversed(class_object.__mro__)
        )
    else:
        mros = reversed(type(class_object).__mro__)
    for base_class in mros:
        all_attributes.update(base_class.__dict__)
    for name, attribute in all_attributes.items():
        if isinstance(attribute, staticmethod):
            all_methods["___s___%s" % name] = inspect.getdoc(attribute)
        elif isinstance(attribute, classmethod):
            all_methods["___c___%s" % name] = inspect.getdoc(attribute)
        elif hasattr(attribute, "__call__"):
            all_methods[name] = inspect.getdoc(attribute)
    return all_methods


def get_canonical_name(name, aliases):
    # We look at the aliases looking for the most specific match first
    base_name = aliases.get(name)
    if base_name is not None:
        return base_name
    for idx in reversed([pos for pos, char in enumerate(name) if char == "."]):
        base_name = aliases.get(name[:idx])
        if base_name is not None:
            return ".".join([base_name, name[idx + 1 :]])
    return name



import inspect
import subprocess
import pickle
import tempfile
import os
import sys
from metaflow import current
from metaflow.plugins.parallel_decorator import ParallelDecorator


class PytorchParallelDecorator(ParallelDecorator):
    name = "pytorch_parallel"
    defaults = {"master_port": None}
    IS_PARALLEL = True

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        return super().task_decorate(
            step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
        )

    def setup_distributed_env(self, flow):
        setup_torch_distributed(self.attributes["master_port"])


def setup_torch_distributed(master_port=None):
    """
    Set up environment variables for PyTorch's distributed (DDP).
    """
    # Choose port depending on run id to reduce probability of collisions, unless
    # provided by the user.
    try:
        master_port = master_port or (51000 + abs(int(current.run_id)) % 10000)
    except:
        # if `int()` fails, i.e. `run_id` is not an `int`, use just a constant port. Can't use `hash()`,
        # as that is not constant.
        master_port = 51001
    os.environ["MASTER_PORT"] = str(master_port)
    os.environ["MASTER_ADDR"] = current.parallel.main_ip
    os.environ["NODE_RANK"] = str(current.parallel.node_index)
    os.environ["WORLD_SIZE"] = str(current.parallel.num_nodes)
    os.environ["NUM_NODES"] = str(current.parallel.num_nodes)
    # Specific for PyTorch Lightning
    os.environ["PL_TORCH_DISTRIBUTED_BACKEND"] = "gloo"  # NCCL crashes on aws batch!



# This file can contain "shortcuts" to other parts of Metaflow (integrations)

# This is an alternative to providing an extension package where you would define
# these aliases in the toplevel file.

# It follows a similar pattern to plugins so that the these integration aliases can be
# turned on and off and avoid exposing things that are not necessarily needed/wanted.

from metaflow.extension_support.integrations import process_integration_aliases

# To enable an alias `metaflow.integrations.get_s3_client` to
# `metaflow.plugins.aws.aws_client.get_aws_client`, use the following:
#
# ALIASES_DESC = [("get_s3_client", ".plugins.aws.aws_client.get_aws_client")]
#
# ALIASES_DESC is a list of tuples:
#  - name: name of the integration alias
#  - obj: object it points to
#
ALIASES_DESC = [("ArgoEvent", ".plugins.argo.argo_events.ArgoEvent")]

# Aliases can be enabled or disabled through configuration or extensions:
#  - ENABLED_INTEGRATION_ALIAS: list of alias names to enable.
#  - TOGGLE_INTEGRATION_ALIAS: if ENABLED_INTEGRATION_ALIAS is not set anywhere
#    (environment variable, configuration or extensions), list of integration aliases
#    to toggle (+<name> or <name> enables, -<name> disables) to build
#    ENABLED_INTEGRATION_ALIAS from the concatenation of the names in
#    ALIASES_DESC (concatenation of the names here as well as in the extensions).

# Keep this line and make sure ALIASES_DESC is above this line.
process_integration_aliases(globals())



import json
import os


from metaflow import current
from metaflow.decorators import StepDecorator
from metaflow.events import Trigger
from metaflow.metadata_provider import MetaDatum
from metaflow.graph import FlowGraph
from metaflow.flowspec import FlowSpec
from .argo_events import ArgoEvent


class ArgoWorkflowsInternalDecorator(StepDecorator):
    name = "argo_workflows_internal"

    defaults = {"auto-emit-argo-events": True}

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        self.task_id = task_id
        self.run_id = run_id

        triggers = []
        # Expose event triggering metadata through current singleton
        for key, payload in os.environ.items():
            if key.startswith("METAFLOW_ARGO_EVENT_PAYLOAD_"):
                if payload != "null":  # Argo-Workflow's None
                    try:
                        payload = json.loads(payload)
                    except (TypeError, ValueError):
                        # There could be arbitrary events that Metaflow doesn't know of
                        payload = {}
                    triggers.append(
                        {
                            "timestamp": payload.get("timestamp"),
                            "id": payload.get("id"),
                            "name": payload.get("name"),  # will exist since filter
                            "type": key[len("METAFLOW_ARGO_EVENT_PAYLOAD_") :].split(
                                "_", 1
                            )[
                                0
                            ],  # infer type from env var key
                            # Add more event metadata here in the future
                        }
                    )

        meta = {}
        if triggers:
            # Enable current.trigger
            current._update_env({"trigger": Trigger(triggers)})
            # Luckily there aren't many events for us to be concerned about the
            # size of the metadata field yet! However we don't really need this
            # metadata outside of the start step so we can save a few bytes in the
            # db.
            if step_name == "start":
                meta["execution-triggers"] = json.dumps(triggers)

        meta["argo-workflow-template"] = os.environ["ARGO_WORKFLOW_TEMPLATE"]
        meta["argo-workflow-name"] = os.environ["ARGO_WORKFLOW_NAME"]
        meta["argo-workflow-namespace"] = os.environ["ARGO_WORKFLOW_NAMESPACE"]
        meta["auto-emit-argo-events"] = self.attributes["auto-emit-argo-events"]
        meta["argo-workflow-template-owner"] = os.environ["METAFLOW_OWNER"]
        entries = [
            MetaDatum(
                field=k, value=v, type=k, tags=["attempt_id:{0}".format(retry_count)]
            )
            for k, v in meta.items()
        ]
        # Register book-keeping metadata for debugging.
        metadata.register_metadata(run_id, step_name, task_id, entries)

    def task_finished(
        self,
        step_name,
        flow: FlowSpec,
        graph: FlowGraph,
        is_task_ok,
        retry_count,
        max_user_code_retries,
    ):
        if not is_task_ok:
            # The task finished with an exception - execution won't
            # continue so no need to do anything here.
            return

        # For `foreach`s, we need to dump the cardinality of the fanout
        # into a file so that Argo Workflows can properly configure
        # the subsequent fanout task via an Output parameter
        #
        # Docker and PNS workflow executors can get output parameters from the base
        # layer (e.g. /tmp), but the Kubelet nor the K8SAPI nor the emissary executors
        # can. It is also unlikely we can get output parameters from the base layer if
        # we run pods with a security context. We work around this constraint by
        # mounting an emptyDir volume.
        if graph[step_name].type == "foreach":
            if graph[step_name].parallel_foreach:
                # If a node is marked as a `parallel_foreach`, pass down the value of
                # `num_parallel` to the subsequent steps.
                with open("/mnt/out/num_parallel", "w") as f:
                    json.dump(flow._parallel_ubf_iter.num_parallel, f)
                # Set splits to 1 since parallelism is handled by JobSet.
                flow._foreach_num_splits = 1
                with open("/mnt/out/task_id_entropy", "w") as file:
                    import uuid

                    file.write(uuid.uuid4().hex[:6])

            with open("/mnt/out/splits", "w") as file:
                json.dump(list(range(flow._foreach_num_splits)), file)
            with open("/mnt/out/split_cardinality", "w") as file:
                json.dump(flow._foreach_num_splits, file)

        # For conditional branches we need to record the value of the switch to disk, in order to pass it as an
        # output from the switching step to be used further down the DAG
        if graph[step_name].type == "split-switch":
            # TODO: A nicer way to access the chosen step?
            _out_funcs, _ = flow._transition
            chosen_step = _out_funcs[0]
            with open("/mnt/out/switch_step", "w") as file:
                file.write(chosen_step)

        # For steps that have a `@parallel` decorator set to them, we will be relying on Jobsets
        # to run the task. In this case, we cannot set anything in the
        # `/mnt/out` directory, since such form of output mounts are not available to Jobset executions.
        if not graph[step_name].parallel_step:
            # Unfortunately, we can't always use pod names as task-ids since the pod names
            # are not static across retries. We write the task-id to a file that is read
            # by the next task here.
            with open("/mnt/out/task_id", "w") as file:
                file.write(self.task_id)

        # Emit Argo Events given that the flow has succeeded. Given that we only
        # emit events when the task succeeds, we can piggy back on this decorator
        # hook which is guaranteed to execute only after rest of the task has
        # finished execution.

        if self.attributes["auto-emit-argo-events"]:
            # Event name is set to metaflow.project.branch.step so that users can
            # place explicit dependencies on namespaced events. Also, argo events
            # sensors don't allow for filtering against absent fields - which limits
            # our ability to subset non-project namespaced events.
            # TODO: Check length limits for fields in Argo Events
            event = ArgoEvent(
                name="metaflow.%s.%s"
                % (current.get("project_flow_name", flow.name), step_name)
            )
            # There should only be one event generated even when the task is retried.
            # Take care to only add to the list and not modify existing values.
            event.add_to_payload("id", current.pathspec)
            event.add_to_payload("pathspec", current.pathspec)
            event.add_to_payload("flow_name", flow.name)
            event.add_to_payload("run_id", self.run_id)
            event.add_to_payload("step_name", step_name)
            event.add_to_payload("task_id", self.task_id)
            # Add @project decorator related fields. These are used to subset
            # @trigger_on_finish related filters.
            for key in (
                "project_name",
                "branch_name",
                "is_user_branch",
                "is_production",
                "project_flow_name",
            ):
                if current.get(key):
                    event.add_to_payload(key, current.get(key))
            # Add more fields here...
            event.add_to_payload("auto-generated-by-metaflow", True)
            # Keep in mind that any errors raised here will fail the run but the task
            # will still be marked as success. That's why we explicitly swallow any
            # errors and instead print them to std.err.
            event.safe_publish(ignore_errors=True)



import importlib
import traceback

from metaflow.metaflow_config_funcs import from_conf

from . import _ext_debug, alias_submodules, get_modules, lazy_load_aliases


def process_plugins(module_globals):
    _resolve_relative_paths(module_globals)
    # Set ENABLED_ and _TOGGLE_ variables. The ENABLED_* variables are read from
    # configuration and the _TOGGLE_* variables are initialized to empty lists to be
    # appended to from the extensions.
    for plugin_category in _plugin_categories:
        upper_category = plugin_category.upper()
        globals()["ENABLED_%s" % upper_category] = from_conf(
            "ENABLED_%s" % upper_category
        )
        globals()["_TOGGLE_%s" % upper_category] = []

        # Initialize the list of available plugins to what is available in Metaflow core
        globals()[_list_for_category(plugin_category)] = _get_ext_plugins(
            module_globals, plugin_category
        )

    try:
        modules_to_import = get_modules("plugins")
        # This is like multiload_all but we load globals independently since we just care
        # about the TOGGLE and ENABLED values
        for m in modules_to_import:
            lazy_load_aliases(
                alias_submodules(m.module, m.tl_package, "plugins", extra_indent=True)
            )
            for n, o in m.module.__dict__.items():
                if n.startswith("TOGGLE_") and n[7:].lower() in _plugin_categories:
                    # Extensions append to the TOGGLE list
                    globals()["_TOGGLE_%s" % n[7:]].extend(o)
                elif n.startswith("ENABLED_") and n[8:].lower() in _plugin_categories:
                    # Extensions override the ENABLED_ setting.
                    globals()[n] = o

            _resolve_relative_paths(m.module.__dict__)
            for plugin_category in _plugin_categories:
                # Collect all the plugins present
                globals()[_list_for_category(plugin_category)].extend(
                    _get_ext_plugins(m.module.__dict__, plugin_category)
                )
    except Exception as e:
        _ext_debug("\tWARNING: ignoring all plugins due to error during import: %s" % e)
        print(
            "WARNING: Plugins did not load -- ignoring all of them which may not "
            "be what you want: %s" % e
        )
        traceback.print_exc()

    # At this point, we have _all_<category>s populated with all the tuples
    # (name, module_class) from all the plugins in all the extensions (if any)
    # We build a dictionary taking the latest presence for each name (so plugins
    # override metaflow core)
    for plugin_category in _plugin_categories:
        upper_category = plugin_category.upper()
        d = globals()[_dict_for_category(plugin_category)] = {}
        for name, class_path in globals()["_all_%ss" % plugin_category]:
            _ext_debug(
                "    Adding %s '%s' from '%s'" % (plugin_category, name, class_path)
            )
            d[name] = class_path

        # Resolve all the ENABLED_* variables. The rules are the following:
        #  - if ENABLED_* is non None, it means it was either set directly by the user
        #    in a configuration file, on the command line or by an extension. In that case
        #    we honor those wishes and completely ignore the extensions' toggles.
        #  - if ENABLED_* is None, we populate it with everything included here and in
        #    all the extensions and use the TOGGLE_ list to produce the final list.
        # The rationale behind this is to support both a configuration option where the
        # plugins enabled are explicitly listed (typical in a lot of software) but also to
        # support a "configuration-less" version where the installation of the extensions
        # determines what is activated.
        if globals()["ENABLED_%s" % upper_category] is None:
            globals()["ENABLED_%s" % upper_category] = (
                list(d) + globals()["_TOGGLE_%s" % upper_category]
            )


def merge_lists(base, overrides, attr):
    # Merge two lists of classes by comparing them for equality using 'attr'.
    # This function prefers anything in 'overrides'. In other words, if a class
    # is present in overrides and matches (according to the equality criterion) a class in
    # base, it will be used instead of the one in base.
    l = list(overrides)
    existing = set([getattr(o, attr) for o in overrides])
    l.extend([d for d in base if getattr(d, attr) not in existing])
    base[:] = l[:]


def get_plugin(category, class_path, name):
    path, cls_name = class_path.rsplit(".", 1)
    try:
        plugin_module = importlib.import_module(path)
    except ImportError as e:
        raise ValueError(
            "Cannot locate %s plugin '%s' at '%s'" % (category, name, path)
        ) from e
    cls = getattr(plugin_module, cls_name, None)
    if cls is None:
        raise ValueError(
            "Cannot locate '%s' class for %s plugin at '%s'"
            % (cls_name, category, path)
        )
    extracted_name = get_plugin_name(category, cls)
    if extracted_name and extracted_name != name:
        raise ValueError(
            "Class '%s' at '%s' for %s plugin expected to be named '%s' but got '%s'"
            % (cls_name, path, category, name, extracted_name)
        )
    globals()[cls_name] = cls
    _ext_debug("        Added %s plugin '%s' from '%s'" % (category, name, class_path))
    return cls


def resolve_plugins(category, path_only=False):
    # Called to return a list of classes that are the available plugins for 'category'

    # The ENABLED_<category> variable is set in process_plugins
    # based on all the plugins that are found; it can contain either names of
    # plugins or -/+<name_of_plugin> indicating a "toggle" to activate/de-activate
    # a plugin.
    list_of_plugins = globals()["ENABLED_%s" % category.upper()]
    _ext_debug("    Resolving %s plugins" % category)
    _ext_debug("        Raw list of plugins is: %s" % str(list_of_plugins))
    set_of_plugins = set()
    for p in list_of_plugins:
        if p.startswith("-"):
            set_of_plugins.discard(p[1:])
        elif p.startswith("+"):
            set_of_plugins.add(p[1:])
        else:
            set_of_plugins.add(p)

    available_plugins = globals()[_dict_for_category(category)]
    name_extractor = _plugin_categories[category]
    if path_only or not name_extractor:
        # If we have no name function, it means we just use the name in the dictionary
        # and we return a dictionary. This is for sidecars mostly as they do not have
        # a field that indicates their name
        to_return = {}
    else:
        to_return = []
    _ext_debug("        Resolved list of plugins is: %s" % str(set_of_plugins))
    # Various error checks to make sure the plugin exists -- basically converts a string
    # representing a class path to the actual class. We try to give useful messages
    # in case of errors.
    for name in set_of_plugins:
        class_path = available_plugins.get(name, None)
        if class_path is None:
            raise ValueError(
                "Configuration requested %s plugin '%s' but no such plugin is available"
                % (category, name)
            )
        if path_only:
            to_return[name] = class_path
        else:
            if name_extractor is not None:
                to_return.append(get_plugin(category, class_path, name))
            else:
                to_return[name] = get_plugin(category, class_path, name)

    return to_return


# Some plugins do not have a field in them indicating their name.
# This is the case for sidecars.
# All other plugins contain a field that indicates their name.
# _plugin_categories contains all the types of plugins and, for ones that have
# a field indicating their name,
# an additional function indicating how to extract the name of the plugin is provided.

# key is the type of plugin
# value is either:
#  - a function to extract the name of the plugin from the plugin itself
#  - None if this is a plugin with no field for its name
_plugin_categories = {
    "step_decorator": lambda x: x.name,
    "flow_decorator": lambda x: x.name,
    "environment": lambda x: x.TYPE,
    "metadata_provider": lambda x: x.TYPE,
    "datastore": lambda x: x.TYPE,
    "dataclient": lambda x: x.TYPE,
    "secrets_provider": lambda x: x.TYPE,
    "gcp_client_provider": lambda x: x.name,
    "deployer_impl_provider": lambda x: x.TYPE,
    "azure_client_provider": lambda x: x.name,
    "sidecar": None,
    "logging_sidecar": None,
    "monitor_sidecar": None,
    "aws_client_provider": lambda x: x.name,
    "cli": lambda x: (
        list(x.commands)[0] if len(x.commands) == 1 else "too many commands"
    ),
    "runner_cli": lambda x: x.name,
    "tl_plugin": None,
}


def get_plugin_name(category, plugin):
    extractor = _plugin_categories[category]
    if extractor:
        return extractor(plugin)
    return None


def _list_for_category(category):
    # Convenience function to name the variable containing List[Tuple[str, str]] where
    # each tuple contains:
    #  - the name of the plugin
    #  - the classpath of the plugin
    return "_all_%ss" % category


def _dict_for_category(category):
    # Convenience function to name the variable containing the same thing as
    # _list_for_category except that it is now in dict form where the key is the name
    # of the plugin
    return "_all_%ss_dict" % category


def _get_ext_plugins(module_globals, category):
    # Convenience function to get the list of Tuple[str, str] describing the plugins
    # available from the extension. This defaults to [] so not all plugins need to be
    # listed.
    return module_globals.get("%sS_DESC" % category.upper(), [])


def _set_ext_plugins(module_globals, category, val):
    module_globals["%sS_DESC" % category.upper()] = val


def _resolve_relative_paths(module_globals):
    # We want to modify all the relevant lists so that the relative paths
    # are made fully qualified paths for the modules
    pkg_path = module_globals["__package__"]
    pkg_components = pkg_path.split(".")

    def resolve_path(class_path):
        # Converts a relative class_path to an absolute one considering that the
        # relative class_path is present in a package pkg_path
        if class_path[0] == ".":
            i = 1
            # Check for multiple "." at the start of the class_path
            while class_path[i] == ".":
                i += 1
            if i > len(pkg_components):
                raise ValueError(
                    "Path '%s' exits out of Metaflow module at %s"
                    % (class_path, pkg_path)
                )
            return (
                ".".join(pkg_components[: -i + 1] if i > 1 else pkg_components)
                + class_path[i - 1 :]
            )
        return class_path

    for plugin_category in _plugin_categories:
        _set_ext_plugins(
            module_globals,
            plugin_category,
            list(
                map(
                    lambda p: (p[0], resolve_path(p[1])),
                    _get_ext_plugins(module_globals, plugin_category),
                )
            ),
        )



import json
import os
import re
import time
from collections import namedtuple
from itertools import chain

from typing import List
from metaflow.exception import MetaflowInternalError, MetaflowTaggingError
from metaflow.tagging_util import validate_tag
from metaflow.util import get_username, resolve_identity_as_tuple, is_stringish

DataArtifact = namedtuple("DataArtifact", "name ds_type ds_root url type sha")

MetaDatum = namedtuple("MetaDatum", "field value type tags")

attempt_id_re = re.compile(r"attempt_id:([0-9]+)")


class MetadataProviderMeta(type):
    def __new__(metaname, classname, bases, attrs):
        return type.__new__(metaname, classname, bases, attrs)

    def _get_info(classobject):
        if not classobject._INFO:
            classobject._INFO = classobject.default_info()
        return classobject._INFO

    def _set_info(classobject, val):
        v = classobject.compute_info(val)
        classobject._INFO = v

    def __init__(classobject, classname, bases, attrs):
        classobject._INFO = None

    INFO = property(_get_info, _set_info)


# From https://stackoverflow.com/questions/22409430/portable-meta-class-between-python2-and-python3
def with_metaclass(mcls):
    def decorator(cls):
        body = vars(cls).copy()
        # clean out class body
        body.pop("__dict__", None)
        body.pop("__weakref__", None)
        return mcls(cls.__name__, cls.__bases__, body)

    return decorator


class ObjectOrder:
    # Consider this list a constant that should never change.
    # Lots of code depend on the membership of this list as
    # well as exact ordering
    _order_as_list = [
        "root",
        "flow",
        "run",
        "step",
        "task",
        "artifact",
        "metadata",
        "self",
    ]
    _order_as_dict = {v: i for i, v in enumerate(_order_as_list)}

    @staticmethod
    def order_to_type(order):
        if order < len(ObjectOrder._order_as_list):
            return ObjectOrder._order_as_list[order]
        return None

    @staticmethod
    def type_to_order(obj_type):
        return ObjectOrder._order_as_dict.get(obj_type)


@with_metaclass(MetadataProviderMeta)
class MetadataProvider(object):
    TYPE = None

    @classmethod
    def metadata_str(cls):
        return "%s@%s" % (cls.TYPE, cls.INFO)

    @classmethod
    def compute_info(cls, val):
        """
        Compute the new information for this provider

        The computed value should be returned and will then be accessible directly as cls.INFO.
        This information will be printed by the client when describing this metadata provider

        Parameters
        ----------
        val : str
            Provider specific information used in computing the new information. For example, this
            can be a path.

        Returns
        -------
        str :
            Value to be set to INFO
        """
        return ""

    @classmethod
    def default_info(cls):
        """
        Returns the default information for this provider

        This should compute and return the default value for the information regarding this provider.
        For example, this can compute where the metadata is stored

        Returns
        -------
        str
            Value to be set by default in INFO
        """
        return ""

    def version(self):
        """
        Returns the version of this provider

        Returns
        -------
        str
            Version of the provider
        """
        return ""

    def new_run_id(self, tags=None, sys_tags=None):
        """
        Creates an ID and registers this new run.

        The run ID will be unique within a given flow.

        Parameters
        ----------
        tags : list, optional
            Tags to apply to this particular run, by default None
        sys_tags : list, optional
            System tags to apply to this particular run, by default None

        Returns
        -------
        int
            Run ID for the run
        """
        raise NotImplementedError()

    def register_run_id(self, run_id, tags=None, sys_tags=None):
        """
        No-op operation in this implementation.

        Parameters
        ----------
        run_id : int
            Run ID for this run
        tags : list, optional
            Tags to apply to this particular run, by default None
        sys_tags : list, optional
            System tags to apply to this particular run, by default None
        Returns
        -------
        bool
            True if a new run was registered; False if it already existed
        """
        raise NotImplementedError()

    def new_task_id(self, run_id, step_name, tags=None, sys_tags=None):
        """
        Creates an ID and registers this new task.

        The task ID will be unique within a flow, run and step

        Parameters
        ----------
        run_id : int
            ID of the run
        step_name : string
            Name of the step
        tags : list, optional
            Tags to apply to this particular task, by default None
        sys_tags : list, optional
            System tags to apply to this particular task, by default None

        Returns
        -------
        int
            Task ID for the task
        """
        raise NotImplementedError()

    def register_task_id(
        self, run_id, step_name, task_id, attempt=0, tags=None, sys_tags=None
    ):
        """
        No-op operation in this implementation.

        Parameters
        ----------
        run_id : int or convertible to int
            Run ID for this run
        step_name : string
            Name of the step
        task_id : int
            Task ID
        tags : list, optional
            Tags to apply to this particular run, by default []
        sys_tags : list, optional
            System tags to apply to this particular run, by default []
        Returns
        -------
        bool
            True if a new run was registered; False if it already existed
        """
        raise NotImplementedError()

    def get_runtime_environment(self, runtime_name):
        """
        Returns a dictionary of environment variables to be set

        Parameters
        ----------
        runtime_name : string
            Name of the runtime for which to get the environment

        Returns
        -------
        dict[string] -> string
            Environment variables from this metadata provider
        """
        return {"METAFLOW_RUNTIME_NAME": runtime_name, "USER": get_username()}

    def register_data_artifacts(
        self, run_id, step_name, task_id, attempt_id, artifacts
    ):
        """
        Registers the fact that the data-artifacts are associated with
        the particular task.

        Artifacts produced by a given task can be associated with the
        task using this call

        Parameters
        ----------
        run_id : int
            Run ID for the task
        step_name : string
            Step name for the task
        task_id : int
            Task ID for the task
        attempt_id : int
            Attempt for the task
        artifacts : List of DataArtifact
            Artifacts associated with this task
        """
        raise NotImplementedError()

    def register_metadata(self, run_id, step_name, task_id, metadata):
        """
        Registers metadata with a task.

        Note that the same metadata can be registered multiple times for the same task (for example
        by multiple attempts). Internally, the timestamp of when the registration call is made is
        also recorded allowing the user to determine the latest value of the metadata.

        Parameters
        ----------
        run_id : int
            Run ID for the task
        step_name : string
            Step name for the task
        task_id : int
            Task ID for the task
        metadata : List of MetaDatum
            Metadata associated with this task
        """
        raise NotImplementedError()

    def start_task_heartbeat(self, flow_id, run_id, step_name, task_id):
        pass

    def start_run_heartbeat(self, flow_id, run_id):
        pass

    def stop_heartbeat(self):
        pass

    @classmethod
    def _get_object_internal(
        cls, obj_type, obj_order, sub_type, sub_order, filters, attempt, *args
    ):
        """
        Return objects for the implementation of this class

        See get_object_internal for the description of what this function does

        Parameters
        ----------
        obj_type : string
            One of 'root', 'flow', 'run', 'step', 'task', 'artifact'
        obj_order: int
            Order in the list ['root', 'flow', 'run', 'step', 'task', 'artifact']
        sub_type : string
            Same as obj_type with the addition of 'metadata', 'self'
        sub_order:
            Order in the same list as the one for obj_order + ['metadata', 'self']
        filters : dict
            Dictionary with keys 'any_tags', 'tags' and 'system_tags'. If specified
            will return only objects that have the specified tags present. Filters
            are ANDed together so all tags must be present for the object to be returned.
        attempt : int or None
            If None, returns artifacts for latest *done* attempt and all metadata. Otherwise,
            returns artifacts for that attempt (existent, done or not) and *all* metadata
            NOTE: Unlike its external facing `get_object`, this function should
            return *all* metadata; the base class will properly implement the
            filter. For artifacts, this function should filter artifacts at
            the backend level.

        Return
        ------
            object or list :
                Depending on the call, the type of object return varies
        """
        raise NotImplementedError()

    def add_sticky_tags(self, tags=None, sys_tags=None):
        """
        Adds tags to be added to every run and task

        Tags can be added to record information about a run/task. Such tags can be specified on a
        per run or task basis using the new_run_id/register_run_id or new_task_id/register_task_id
        functions but can also be set globally using this function. Tags added here will be
        added to every run/task created after this call is made.

        Parameters
        ----------
        tags : list, optional
            Tags to add to every run/task, by default None
        sys_tags : list, optional
            System tags to add to every run/task, by default None
        """
        if tags:
            self.sticky_tags.update(tags)
        if sys_tags:
            self.sticky_sys_tags.update(sys_tags)

    @classmethod
    def get_object(cls, obj_type, sub_type, filters, attempt, *args):
        """Returns the requested object depending on obj_type and sub_type

        obj_type can be one of 'root', 'flow', 'run', 'step', 'task',
        or 'artifact'

        sub_type describes the aggregation required and can be either:
        'metadata', 'self' or any of obj_type provided that it is slotted below
        the object itself. For example, if obj_type is 'flow', you can
        specify 'run' to get all the runs in that flow.
        A few special rules:
            - 'metadata' is only allowed for obj_type 'task'
            - For obj_type 'artifact', only 'self' is allowed
        A few examples:
            - To get a list of all flows:
                - set obj_type to 'root' and sub_type to 'flow'
            - To get a list of all tasks:
                - set obj_type to 'root' and sub_type to 'task'
            - To get a list of all artifacts in a task:
                - set obj_type to 'task' and sub_type to 'artifact'
            - To get information about a specific flow:
                - set obj_type to 'flow' and sub_type to 'self'

        Parameters
        ----------
        obj_type : string
            One of 'root', 'flow', 'run', 'step', 'task', 'artifact' or 'metadata'
        sub_type : string
            Same as obj_type with the addition of 'self'
        filters : dict
            Dictionary with keys 'any_tags', 'tags' and 'system_tags'. If specified
            will return only objects that have the specified tags present. Filters
            are ANDed together so all tags must be present for the object to be returned.
        attempt : int or None
            If None, for metadata and artifacts:
              - returns information about the latest attempt for artifacts
              - returns all metadata across all attempts
            Otherwise, returns information about metadata and artifacts for that
            attempt only.
            NOTE: For older versions of Metaflow (pre 2.4.0), the attempt for
            metadata is not known; in that case, all metadata is returned (as
            if None was passed in).

        Return
        ------
            object or list :
                Depending on the call, the type of object return varies
        """
        type_order = ObjectOrder.type_to_order(obj_type)
        sub_order = ObjectOrder.type_to_order(sub_type)

        if type_order is None:
            raise MetaflowInternalError(msg="Cannot find type %s" % obj_type)
        if type_order >= ObjectOrder.type_to_order("metadata"):
            raise MetaflowInternalError(msg="Type %s is not allowed" % obj_type)

        if sub_order is None:
            raise MetaflowInternalError(msg="Cannot find subtype %s" % sub_type)

        if type_order >= sub_order:
            raise MetaflowInternalError(
                msg="Subtype %s not allowed for %s" % (sub_type, obj_type)
            )

        # Metadata is always only at the task level
        if sub_type == "metadata" and obj_type != "task":
            raise MetaflowInternalError(
                msg="Metadata can only be retrieved at the task level"
            )

        if attempt is not None:
            try:
                attempt_int = int(attempt)
                if attempt_int < 0:
                    raise ValueError("Attempt can only be positive")
            except ValueError:
                raise ValueError("Attempt can only be a positive integer")
        else:
            attempt_int = None

        pre_filter = cls._get_object_internal(
            obj_type, type_order, sub_type, sub_order, filters, attempt_int, *args
        )
        if attempt_int is None or sub_type != "metadata":
            # If no attempt or not for metadata, just return as is
            return pre_filter
        return MetadataProvider._reconstruct_metadata_for_attempt(
            pre_filter, attempt_int
        )

    @classmethod
    def mutate_user_tags_for_run(
        cls, flow_id, run_id, tags_to_remove=None, tags_to_add=None
    ):
        """
        Mutate the set of user tags for a run.

        Removals logically get applied after additions.  Operations occur as a batch atomically.
        Parameters
        ----------
        flow_id : str
            Flow id, that the run belongs to.
        run_id: str
            Run id, together with flow_id, that identifies the specific Run whose tags to mutate
        tags_to_remove: iterable over str
            Iterable over tags to remove
        tags_to_add: iterable over str
            Iterable over tags to add

        Return
        ------
        Run tags after mutation operations
        """
        # perform common validation, across all provider implementations
        if tags_to_remove is None:
            tags_to_remove = []
        if tags_to_add is None:
            tags_to_add = []
        if not tags_to_add and not tags_to_remove:
            raise MetaflowTaggingError("Must add or remove at least one tag")

        if is_stringish(tags_to_add):
            raise MetaflowTaggingError("tags_to_add may not be a string")

        if is_stringish(tags_to_remove):
            raise MetaflowTaggingError("tags_to_remove may not be a string")

        def _is_iterable(something):
            try:
                iter(something)
                return True
            except TypeError:
                return False

        if not _is_iterable(tags_to_add):
            raise MetaflowTaggingError("tags_to_add must be iterable")
        if not _is_iterable(tags_to_remove):
            raise MetaflowTaggingError("tags_to_remove must be iterable")

        # check each tag is valid
        for tag in chain(tags_to_add, tags_to_remove):
            validate_tag(tag)

        # onto subclass implementation
        final_user_tags = cls._mutate_user_tags_for_run(
            flow_id, run_id, tags_to_add=tags_to_add, tags_to_remove=tags_to_remove
        )
        return final_user_tags

    @classmethod
    def _mutate_user_tags_for_run(
        cls, flow_id, run_id, tags_to_add=None, tags_to_remove=None
    ):
        """
        To be implemented by subclasses of MetadataProvider.

        See mutate_user_tags_for_run() for expectations.
        """
        raise NotImplementedError()

    def _all_obj_elements(self, tags=None, sys_tags=None):
        return MetadataProvider._all_obj_elements_static(
            self._flow_name, tags=tags, sys_tags=sys_tags
        )

    @staticmethod
    def _all_obj_elements_static(flow_name, tags=None, sys_tags=None):
        user = get_username()
        return {
            "flow_id": flow_name,
            "user_name": user,
            "tags": list(tags) if tags else [],
            "system_tags": list(sys_tags) if sys_tags else [],
            "ts_epoch": int(round(time.time() * 1000)),
        }

    def _flow_to_json(self):
        # No need to store tags, sys_tags or username at the flow level
        # since runs are the top level logical concept, which is where we
        # store tags, sys_tags and username
        return {"flow_id": self._flow_name, "ts_epoch": int(round(time.time() * 1000))}

    def _run_to_json(self, run_id=None, tags=None, sys_tags=None):
        return MetadataProvider._run_to_json_static(
            self._flow_name, run_id=run_id, tags=tags, sys_tags=sys_tags
        )

    @staticmethod
    def _run_to_json_static(flow_name, run_id=None, tags=None, sys_tags=None):
        if run_id is not None:
            d = {"run_number": run_id}
        else:
            d = {}
        d.update(MetadataProvider._all_obj_elements_static(flow_name, tags, sys_tags))
        return d

    def _step_to_json(self, run_id, step_name, tags=None, sys_tags=None):
        d = {"run_number": run_id, "step_name": step_name}
        d.update(self._all_obj_elements(tags, sys_tags))
        return d

    def _task_to_json(self, run_id, step_name, task_id=None, tags=None, sys_tags=None):
        d = {"run_number": run_id, "step_name": step_name}
        if task_id is not None:
            d["task_id"] = task_id
        d.update(self._all_obj_elements(tags, sys_tags))
        return d

    def _object_to_json(
        self,
        obj_type,
        run_id=None,
        step_name=None,
        task_id=None,
        tags=None,
        sys_tags=None,
    ):
        if obj_type == "task":
            return self._task_to_json(run_id, step_name, task_id, tags, sys_tags)
        if obj_type == "step":
            return self._step_to_json(run_id, step_name, tags, sys_tags)
        if obj_type == "run":
            return self._run_to_json(run_id, tags, sys_tags)
        return self._flow_to_json()

    def _artifacts_to_json(self, run_id, step_name, task_id, attempt_id, artifacts):
        result = []
        for art in artifacts:
            d = {
                "run_number": run_id,
                "step_name": step_name,
                "task_id": task_id,
                "attempt_id": attempt_id,
                "name": art.name,
                "content_type": art.type,
                "type": "metaflow.artifact",
                "sha": art.sha,
                "ds_type": art.ds_type,
                "location": art.url if art.url else ":root:%s" % art.ds_root,
            }
            d.update(self._all_obj_elements(self.sticky_tags, self.sticky_sys_tags))
            result.append(d)
        return result

    def _metadata_to_json(self, run_id, step_name, task_id, metadata):
        user = get_username()
        return [
            {
                "flow_id": self._flow_name,
                "run_number": run_id,
                "step_name": step_name,
                "task_id": task_id,
                "field_name": datum.field,
                "type": datum.type,
                "value": datum.value,
                "tags": list(set(datum.tags)) if datum.tags else [],
                "user_name": user,
                "ts_epoch": int(round(time.time() * 1000)),
            }
            for datum in metadata
        ]

    def _get_system_info_as_dict(self):
        """This function drives:
        - sticky system tags initialization
        - task-level metadata generation
        """
        sys_info = dict()
        env = self._environment.get_environment_info()
        sys_info["runtime"] = env["runtime"]
        sys_info["python_version"] = env["python_version_code"]
        identity_type, identity_value = resolve_identity_as_tuple()
        sys_info[identity_type] = identity_value
        if env["metaflow_version"]:
            sys_info["metaflow_version"] = env["metaflow_version"]
        if "metaflow_r_version" in env:
            sys_info["metaflow_r_version"] = env["metaflow_r_version"]
        if "r_version_code" in env:
            sys_info["r_version"] = env["r_version_code"]
        return sys_info

    def _get_git_info_as_dict(self):
        git_info = {}
        # NOTE: For flows executing remotely, we want to read from the INFO file of the code package that contains
        # information on the original environment that deployed the flow.
        # Otherwise git related info will be missing, as the repository is not part of the codepackage.
        from metaflow.packaging_sys import MetaflowCodeContent

        env = MetaflowCodeContent.get_info() or self._environment.get_environment_info()
        for key in [
            "repo_url",
            "branch_name",
            "commit_sha",
            "has_uncommitted_changes",
        ]:
            if key in env and env[key]:
                git_info[key] = env[key]

        return git_info

    def _get_system_tags(self):
        """Convert system info dictionary into a list of system tags"""
        return [
            "{}:{}".format(k, v) for k, v in self._get_system_info_as_dict().items()
        ]

    def _register_system_metadata(self, run_id, step_name, task_id, attempt):
        """Gather up system and code packaging info and register them as task metadata"""
        metadata = []
        # Take everything from system info and store them as metadata
        sys_info = self._get_system_info_as_dict()

        # field, and type could get long in theory...can the metadata backend handle it?
        # E.g. as of 5/9/2022 Metadata service's DB says VARCHAR(255).
        # It is likely overkill to fail a flow over an over-flow. We should expect the
        # backend to try to tolerate this (e.g. enlarge columns, truncation fallback).
        metadata.extend(
            MetaDatum(
                field=str(k),
                value=str(v),
                type=str(k),
                tags=["attempt_id:{0}".format(attempt)],
            )
            for k, v in sys_info.items()
        )
        # Also store code packaging information
        code_sha = os.environ.get("METAFLOW_CODE_SHA")
        if code_sha:
            code_url = os.environ.get("METAFLOW_CODE_URL")
            code_ds = os.environ.get("METAFLOW_CODE_DS")
            code_metadata = os.environ.get("METAFLOW_CODE_METADATA")
            metadata.append(
                MetaDatum(
                    field="code-package",
                    value=json.dumps(
                        {
                            "ds_type": code_ds,
                            "sha": code_sha,
                            "location": code_url,
                            "metadata": code_metadata,
                        }
                    ),
                    type="code-package",
                    tags=["attempt_id:{0}".format(attempt)],
                )
            )
        # Add script name as metadata
        script_name = self._environment.get_environment_info()["script"]
        metadata.append(
            MetaDatum(
                field="script-name",
                value=script_name,
                type="script-name",
                tags=["attempt_id:{0}".format(attempt)],
            )
        )
        # And add git metadata
        git_info = self._get_git_info_as_dict()
        if git_info:
            metadata.append(
                MetaDatum(
                    field="git-info",
                    value=json.dumps(git_info),
                    type="git-info",
                    tags=["attempt_id:{0}".format(attempt)],
                )
            )
        if metadata:
            self.register_metadata(run_id, step_name, task_id, metadata)

    @classmethod
    def filter_tasks_by_metadata(
        cls,
        flow_name: str,
        run_id: str,
        step_name: str,
        field_name: str,
        pattern: str,
    ) -> List[str]:
        """
        Filter tasks by metadata field and pattern, returning task pathspecs that match criteria.

        Parameters
        ----------
        flow_name : str
            Flow name, that the run belongs to.
        run_id: str
            Run id, together with flow_id, that identifies the specific Run whose tasks to query
        step_name: str
            Step name to query tasks from
        field_name: str
            Metadata field name to query
        pattern: str
            Pattern to match in metadata field value

        Returns
        -------
        List[str]
            List of task pathspecs that satisfy the query
        """
        raise NotImplementedError()

    @staticmethod
    def _apply_filter(elts, filters):
        if filters is None:
            return elts
        starting_point = elts
        result = []
        for key, value in filters.items():
            if key == "any_tags":
                for obj in starting_point:
                    if value in obj.get("tags", []) or value in obj.get(
                        "system_tags", []
                    ):
                        result.append(obj)
            if key == "tags":
                for obj in starting_point:
                    if value in obj.get("tags", []):
                        result.append(obj)
            if key == "system_tags":
                for obj in starting_point:
                    if value in obj.get("system_tags", []):
                        result.append(obj)
            starting_point = result
            result = []
        return starting_point

    @staticmethod
    def _reconstruct_metadata_for_attempt(all_metadata, attempt_id):
        have_all_attempt_id = True
        attempts_start = {}
        post_filter = []
        for v in all_metadata:
            if v["field_name"] == "attempt":
                attempts_start[int(v["value"])] = v["ts_epoch"]
            all_tags = v.get("tags")
            if all_tags is None:
                all_tags = []
            for t in all_tags:
                match_result = attempt_id_re.match(t)
                if match_result:
                    if int(match_result.group(1)) == attempt_id:
                        post_filter.append(v)
                    break
            else:
                # We didn't encounter a match for attempt_id
                have_all_attempt_id = False

        if not have_all_attempt_id:
            # We reconstruct base on the attempts_start
            start_ts = attempts_start.get(attempt_id, -1)
            if start_ts < 0:
                return []  # No metadata since the attempt hasn't started
            # Doubt we will be using Python in year 3000
            end_ts = attempts_start.get(attempt_id + 1, 32503680000000)
            post_filter = [
                v
                for v in all_metadata
                if v["ts_epoch"] >= start_ts and v["ts_epoch"] < end_ts
            ]

        return post_filter

    def __init__(self, environment, flow, event_logger, monitor):
        self._task_id_seq = -1
        self.sticky_tags = set()
        self.sticky_sys_tags = set()
        self._flow_name = flow.name
        self._event_logger = event_logger
        self._monitor = monitor
        self._environment = environment
        self._runtime = os.environ.get("METAFLOW_RUNTIME_NAME", "dev")
        self.add_sticky_tags(sys_tags=self._get_system_tags())



import os
import sys
from typing import Dict, Any, Optional, Union


class SystemLogger(object):
    def __init__(self):
        self._logger = None
        self._flow_name = None

    def __del__(self):
        if self._flow_name == "not_a_real_flow":
            self.logger.terminate()

    def init_system_logger(
        self, flow_name: str, logger: "metaflow.event_logger.NullEventLogger"
    ):
        self._flow_name = flow_name
        self._logger = logger

    def _init_logger_outside_flow(self):
        from .system_utils import DummyFlow
        from .system_utils import init_environment_outside_flow
        from metaflow.plugins import LOGGING_SIDECARS
        from metaflow.metaflow_config import DEFAULT_EVENT_LOGGER

        self._flow_name = "not_a_real_flow"
        _flow = DummyFlow(self._flow_name)
        _environment = init_environment_outside_flow(_flow)
        _logger = LOGGING_SIDECARS[DEFAULT_EVENT_LOGGER](_flow, _environment)
        return _logger

    @property
    def logger(self) -> Optional["metaflow.event_logger.NullEventLogger"]:
        if self._logger is None:
            # This happens if the logger is being accessed outside of a flow
            # We start a logger with a dummy flow and a default environment
            self._debug("Started logger outside of a flow")
            self._logger = self._init_logger_outside_flow()
            self._logger.start()
        return self._logger

    @staticmethod
    def _debug(msg: str):
        """
        Log a debug message to stderr.

        Parameters
        ----------
        msg : str
            Message to log.

        """
        if os.environ.get("METAFLOW_DEBUG_SIDECAR", "0").lower() not in (
            "0",
            "false",
            "",
        ):
            print("system logger: %s" % msg, file=sys.stderr)

    def log_event(
        self, level: str, module: str, name: str, payload: Optional[Any] = None
    ):
        """
        Log an event to the event logger.

        Parameters
        ----------
        level : str
            Log level of the event. Can be one of "info", "warning", "error", "critical", "debug".
        module : str
            Module of the event. Usually the name of the class, function, or module that the event is being logged from.
        name : str
            Name of the event. Used to qualify the event type.
        payload : Optional[Any], default None
            Payload of the event. Contains the event data.
        """
        self.logger.log(
            {
                "level": level,
                "module": module,
                "name": name,
                "payload": payload if payload is not None else {},
            }
        )



import atexit
import importlib
import importlib.util
import itertools
import pickle
import re
import sys

from .consts import OP_CALLFUNC, OP_GETVAL, OP_SETVAL
from .client import Client
from .utils import get_canonical_name


def _clean_client(client):
    client.cleanup()


class _WrappedModule(object):
    def __init__(self, loader, prefix, exports, client):
        self._loader = loader
        self._prefix = prefix
        self._client = client
        is_match = re.compile(
            r"^%s\.([a-zA-Z_][a-zA-Z0-9_]*)$" % prefix.replace(".", r"\.")  # noqa W605
        )
        self._exports = {}
        self._aliases = exports.get("aliases", [])
        for k in ("classes", "functions", "values"):
            result = []
            for item in exports.get(k, []):
                m = is_match.match(item)
                if m:
                    result.append(m.group(1))
            self._exports[k] = result
        result = []
        for item, _ in exports.get("exceptions", []):
            m = is_match.match(item)
            if m:
                result.append(m.group(1))
        self._exports["exceptions"] = result

    def __getattr__(self, name):
        if name == "__loader__":
            return self._loader
        if name == "__spec__":
            return importlib.util.spec_from_loader(self._prefix, self._loader)
        if name in ("__name__", "__package__"):
            return self._prefix
        if name in ("__file__", "__path__"):
            return self._client.name

        # Make the name canonical because the prefix is also canonical.
        name = get_canonical_name(self._prefix + "." + name, self._aliases)[
            len(self._prefix) + 1 :
        ]
        if name in self._exports["classes"] or name in self._exports["exceptions"]:
            # We load classes and exceptions lazily
            return self._client.get_local_class("%s.%s" % (self._prefix, name))
        elif name in self._exports["functions"]:
            # TODO: Grab doc back from the remote side like in _make_method
            def func(*args, **kwargs):
                return self._client.stub_request(
                    None, OP_CALLFUNC, "%s.%s" % (self._prefix, name), *args, **kwargs
                )

            func.__name__ = name
            func.__doc__ = "Unknown (TODO)"
            return func
        elif name in self._exports["values"]:
            return self._client.stub_request(
                None, OP_GETVAL, "%s.%s" % (self._prefix, name)
            )
        else:
            # Try to see if this is a submodule that we can load
            m = None
            try:
                submodule_name = ".".join([self._prefix, name])
                m = importlib.import_module(submodule_name)
            except ImportError:
                pass
            if m is None:
                raise AttributeError(
                    "module '%s' has no attribute '%s' -- contact the author of the "
                    "configuration if this is something "
                    "you expect to work (support may be added if it exists in the "
                    "original library)" % (self._prefix, name)
                )
            return m

    def __setattr__(self, name, value):
        if name in (
            "package",
            "__spec__",
            "_loader",
            "_prefix",
            "_client",
            "_exports",
            "_exception_classes",
            "_aliases",
        ):
            object.__setattr__(self, name, value)
            return
        if isinstance(value, _WrappedModule):
            # This is when a module sets itself as an attribute of another
            # module when loading
            object.__setattr__(self, name, value)
            return

        # Make the name canonical because the prefix is also canonical.
        name = get_canonical_name(self._prefix + "." + name, self._aliases)[
            len(self._prefix) + 1 :
        ]
        if name in self._exports["values"]:
            self._client.stub_request(
                None, OP_SETVAL, "%s.%s" % (self._prefix, name), value
            )
        elif name in self._exports["classes"] or name in self._exports["functions"]:
            raise ValueError
        else:
            raise AttributeError(name)


class ModuleImporter(object):
    """
    A custom import hook that proxies module imports to a different Python environment.

    This class implements the MetaPathFinder and Loader protocols (PEP 451) to enable
    "environment escape" - allowing the current Python process to import and use modules
    from a different Python interpreter with potentially different versions or packages.

    When a module is imported through this importer:
    1. A client spawns a server process in the target Python environment
    2. The module is loaded in the remote environment
    3. A _WrappedModule proxy is returned that forwards all operations (function calls,
       attribute access, etc.) to the remote environment via RPC
    4. Data is serialized/deserialized using pickle for cross-environment communication

    Args:
        python_executable: Path to the Python interpreter for the remote environment
        pythonpath: Python path to use in the remote environment
        max_pickle_version: Maximum pickle protocol version supported by remote interpreter
        config_dir: Directory containing configuration for the environment escape
        module_prefixes: List of module name prefixes to handle
    """

    def __init__(
        self,
        python_executable,
        pythonpath,
        max_pickle_version,
        config_dir,
        module_prefixes,
    ):
        self._module_prefixes = module_prefixes
        self._python_executable = python_executable
        self._pythonpath = pythonpath
        self._config_dir = config_dir
        self._client = None
        self._max_pickle_version = max_pickle_version
        self._handled_modules = None
        self._aliases = {}

    def find_spec(self, fullname, path=None, target=None):
        if self._handled_modules is not None:
            if get_canonical_name(fullname, self._aliases) in self._handled_modules:
                return importlib.util.spec_from_loader(fullname, self)
            return None
        if any([fullname.startswith(prefix) for prefix in self._module_prefixes]):
            # We potentially handle this
            return importlib.util.spec_from_loader(fullname, self)
        return None

    def create_module(self, spec):
        # Return the pre-created wrapped module for this spec
        self._initialize_client()

        fullname = spec.name
        canonical_fullname = get_canonical_name(fullname, self._aliases)
        # Modules are created canonically but we need to handle any of the aliases.
        wrapped_module = self._handled_modules.get(canonical_fullname)
        if wrapped_module is None:
            raise ImportError(f"No module named '{fullname}'")
        return wrapped_module

    def exec_module(self, module):
        # No initialization needed since the wrapped module returned by
        # create_module() is fully initialized
        pass

    def _initialize_client(self):
        if self._client is not None:
            return

        # We initialize a client and query the modules we handle
        # The max_pickle_version is the pickle version that the server (so
        # the underlying interpreter we call into) supports; we determine
        # what version the current environment support and take the minimum
        # of those two
        max_pickle_version = min(self._max_pickle_version, pickle.HIGHEST_PROTOCOL)

        self._client = Client(
            self._module_prefixes,
            self._python_executable,
            self._pythonpath,
            max_pickle_version,
            self._config_dir,
        )
        atexit.register(_clean_client, self._client)

        # Get information about overrides and what the server knows about
        exports = self._client.get_exports()

        prefixes = set()
        export_classes = exports.get("classes", [])
        export_functions = exports.get("functions", [])
        export_values = exports.get("values", [])
        export_exceptions = exports.get("exceptions", [])
        self._aliases = exports.get("aliases", {})
        for name in itertools.chain(
            export_classes,
            export_functions,
            export_values,
            (e[0] for e in export_exceptions),
        ):
            splits = name.rsplit(".", 1)
            prefixes.add(splits[0])
        # We will make sure that we create modules even for "empty" prefixes
        # because packages are always loaded hierarchically so if we have
        # something in `a.b.c` but nothing directly in `a`, we still need to
        # create a module named `a`. There is probably a better way of doing this
        all_prefixes = list(prefixes)
        for prefix in all_prefixes:
            parts = prefix.split(".")
            cur = parts[0]
            for i in range(1, len(parts)):
                prefixes.add(cur)
                cur = ".".join([cur, parts[i]])

        # We now know all the modules that we can handle. We update
        # handled_module and return the module if we have it or raise ImportError
        self._handled_modules = {}
        for prefix in prefixes:
            self._handled_modules[prefix] = _WrappedModule(
                self, prefix, exports, self._client
            )


def create_modules(python_executable, pythonpath, max_pickle_version, path, prefixes):
    # This is an extra verification to make sure we are not trying to use the
    # environment escape for something that is in the system
    for prefix in prefixes:
        try:
            importlib.import_module(prefix)
        except ImportError:
            pass
        else:
            # pass
            raise RuntimeError(
                "Trying to override %s when module exists in system" % prefix
            )

    # The first version forces the use of the environment escape even if the module
    # exists in the system. This is useful for testing to make sure that the
    # environment escape is used. The second version is more production friendly and
    # will only use the environment escape if the module cannot be found

    # sys.meta_path.insert(0, ModuleImporter(python_path, path, prefixes))
    sys.meta_path.append(
        ModuleImporter(
            python_executable, pythonpath, max_pickle_version, path, prefixes
        )
    )



import os

from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator
from metaflow.metaflow_config import DEFAULT_SECRETS_ROLE
from metaflow.plugins.secrets.secrets_spec import SecretSpec
from metaflow.plugins.secrets.utils import (
    get_secrets_backend_provider,
    validate_env_vars,
    validate_env_vars_across_secrets,
    validate_env_vars_vs_existing_env,
)
from metaflow.unbounded_foreach import UBF_TASK


class SecretsDecorator(StepDecorator):
    """
    Specifies secrets to be retrieved and injected as environment variables prior to
    the execution of a step.

    Parameters
    ----------
    sources : List[Union[str, Dict[str, Any]]], default: []
        List of secret specs, defining how the secrets are to be retrieved
    role : str, optional, default: None
        Role to use for fetching secrets
    """

    name = "secrets"
    defaults = {
        "sources": [],
        "role": None,
    }

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        if (
            ubf_context
            and ubf_context == UBF_TASK
            and os.environ.get("METAFLOW_RUNTIME_ENVIRONMENT", "local") == "local"
        ):
            # We will skip the secret injection for "locally" launched UBF_TASK (worker) tasks
            # When we "locally" run @parallel tasks, the control task will create the worker tasks and the environment variables
            # of the control task are inherited by the worker tasks. If we don't skip setting secrets in the worker task then the
            # worker tasks will try to set the environment variables again which will cause a clash with the control task's env vars,
            # causing the @secrets' `task_pre_step` to fail. In remote settings, (e.g. AWS Batch/Kubernetes), the worker task and
            # control task are independently created so there is no chances of an env var clash.
            return
        # List of pairs (secret_spec, env_vars_from_this_spec)
        all_secrets_env_vars = []
        secret_specs = []

        # Role (in terms of RBAC) to use when retrieving secrets.
        # This is a general concept applicable to multiple backends
        # E.g in AWS, this would be an IAM Role ARN.
        #
        # Config precedence (decreasing):
        # - Source level: @secrets(source=[{"role": ...}])
        # - Decorator level: @secrets(role=...)
        # - Metaflow config key DEFAULT_SECRETS_ROLE
        role = self.attributes["role"]
        if role is None:
            role = DEFAULT_SECRETS_ROLE

        for secret_spec_str_or_dict in self.attributes["sources"]:
            if isinstance(secret_spec_str_or_dict, str):
                secret_specs.append(
                    SecretSpec.secret_spec_from_str(secret_spec_str_or_dict, role=role)
                )
            elif isinstance(secret_spec_str_or_dict, dict):
                secret_specs.append(
                    SecretSpec.secret_spec_from_dict(secret_spec_str_or_dict, role=role)
                )
            else:
                raise MetaflowException(
                    "@secrets sources items must be either a string or a dict"
                )

        for secret_spec in secret_specs:
            secrets_backend_provider = get_secrets_backend_provider(
                secret_spec.secrets_backend_type
            )
            try:
                env_vars_for_secret = secrets_backend_provider.get_secret_as_dict(
                    secret_spec.secret_id,
                    options=secret_spec.options,
                    role=secret_spec.role,
                )
            except Exception as e:
                raise MetaflowException(
                    "Failed to retrieve secret '%s': %s" % (secret_spec.secret_id, e)
                )
            try:
                validate_env_vars(env_vars_for_secret)
            except ValueError as e:
                raise MetaflowException(
                    "Invalid env vars from secret %s: %s"
                    % (secret_spec.secret_id, str(e))
                )
            all_secrets_env_vars.append((secret_spec, env_vars_for_secret))

        validate_env_vars_across_secrets(all_secrets_env_vars)
        validate_env_vars_vs_existing_env(all_secrets_env_vars)

        # By this point
        # all_secrets_env_vars contains a list of dictionaries... env maps.
        # - env maps must be disjoint from each other
        # - env maps must be disjoint from existing current process os.environ

        for secrets_env_vars in all_secrets_env_vars:
            os.environ.update(secrets_env_vars[1].items())



import os
import re
from metaflow.exception import MetaflowException

DISALLOWED_SECRETS_ENV_VAR_PREFIXES = ["METAFLOW_"]


def get_default_secrets_backend_type():
    from metaflow.metaflow_config import DEFAULT_SECRETS_BACKEND_TYPE

    if DEFAULT_SECRETS_BACKEND_TYPE is None:
        raise MetaflowException(
            "No default secrets backend type configured, but needed by @secrets. "
            "Set METAFLOW_DEFAULT_SECRETS_BACKEND_TYPE."
        )
    return DEFAULT_SECRETS_BACKEND_TYPE


def validate_env_vars_across_secrets(all_secrets_env_vars):
    vars_injected_by = {}
    for secret_spec, env_vars in all_secrets_env_vars:
        for k in env_vars:
            if k in vars_injected_by:
                raise MetaflowException(
                    "Secret '%s' will inject '%s' as env var, and it is also added by '%s'"
                    % (secret_spec, k, vars_injected_by[k])
                )
            vars_injected_by[k] = secret_spec


def validate_env_vars_vs_existing_env(all_secrets_env_vars):
    for secret_spec, env_vars in all_secrets_env_vars:
        for k in env_vars:
            if k in os.environ:
                raise MetaflowException(
                    "Secret '%s' will inject '%s' as env var, but it already exists in env"
                    % (secret_spec, k)
                )


def validate_env_vars(env_vars):
    for k, v in env_vars.items():
        if not isinstance(k, str):
            raise MetaflowException("Found non string key %s (%s)" % (str(k), type(k)))
        if not isinstance(v, str):
            raise MetaflowException(
                "Found non string value %s (%s)" % (str(v), type(v))
            )
        if not re.fullmatch("[a-zA-Z_][a-zA-Z0-9_]*", k):
            raise MetaflowException("Found invalid env var name '%s'." % k)
        for disallowed_prefix in DISALLOWED_SECRETS_ENV_VAR_PREFIXES:
            if k.startswith(disallowed_prefix):
                raise MetaflowException(
                    "Found disallowed env var name '%s' (starts with '%s')."
                    % (k, disallowed_prefix)
                )


def get_secrets_backend_provider(secrets_backend_type):
    from metaflow.plugins import SECRETS_PROVIDERS

    try:
        provider_cls = [
            pc for pc in SECRETS_PROVIDERS if pc.TYPE == secrets_backend_type
        ][0]
        return provider_cls()
    except IndexError:
        raise MetaflowException(
            "Unknown secrets backend type %s (available types: %s)"
            % (
                secrets_backend_type,
                ", ".join(pc.TYPE for pc in SECRETS_PROVIDERS if pc.TYPE != "inline"),
            )
        )



from metaflow.exception import MetaflowTaggingError
from metaflow.util import unicode_type, bytes_type


def is_utf8_encodable(x):
    """
    Returns true if the object can be encoded with UTF-8
    """
    try:
        x.encode("utf-8")
        return True
    except UnicodeError:
        return False


def is_utf8_decodable(x):
    """
    Returns true if the object can be decoded with UTF-8
    """
    try:
        x.decode("utf-8")
        return True
    except UnicodeError:
        return False


# How many user tags are allowed on a run
MAX_USER_TAG_SET_SIZE = 50
# How long may an individual tag value be
MAX_TAG_SIZE = 500


def validate_tags(tags, existing_tags=None):
    """
    Raises MetaflowTaggingError if invalid based on these rules:

    Tag set size is too large. But it's OK if tag set is not larger
    than an existing tag set (if provided).

    Then, we validate each tag.  See validate_tag()
    """
    tag_set = frozenset(tags)
    if len(tag_set) > MAX_USER_TAG_SET_SIZE:
        # We want to allow user to remediate excessively large tag sets via tag mutation
        if existing_tags is None or len(frozenset(existing_tags)) < len(tag_set):
            raise MetaflowTaggingError(
                msg="Cannot increase size of tag set beyond %d"
                % (MAX_USER_TAG_SET_SIZE,)
            )
    for tag in tag_set:
        validate_tag(tag)


def validate_tag(tag):
    """
    - Tag must be either of bytes-type or unicode-type.
    - If tag is of bytes-type, it must be UTF-8 decodable
    - If tag is of unicode-type, it must be UTF-8 encodable
    - Tag may not be empty string.
    - Tag cannot be too long (500 chars)
    """
    if isinstance(tag, bytes_type):
        if not is_utf8_decodable(tag):
            raise MetaflowTaggingError("Tags must be UTF-8 decodable")
    elif isinstance(tag, unicode_type):
        if not is_utf8_encodable(tag):
            raise MetaflowTaggingError("Tags must be UTF-8 encodable")
    else:
        raise MetaflowTaggingError(
            "Tags must be some kind of string (bytes or unicode), got %s",
            str(type(tag)),
        )
    if not len(tag):
        raise MetaflowTaggingError("Tags must not be empty string")
    if len(tag) > MAX_TAG_SIZE:
        raise MetaflowTaggingError("Tag is too long %d > %d" % (len(tag), MAX_TAG_SIZE))



import importlib
import json
import os
import sys

from typing import Any, ClassVar, Dict, Optional, TYPE_CHECKING, Type, List

from metaflow.metaflow_config import CLICK_API_PROCESS_CONFIG

from .subprocess_manager import SubprocessManager
from .utils import get_lower_level_group, handle_timeout, temporary_fifo, with_dir

if TYPE_CHECKING:
    import metaflow.runner.deployer

# NOTE: This file is separate from the deployer.py file to prevent circular imports.
# This file is needed in any of the DeployerImpl implementations
# (like argo_workflows_deployer.py) which is in turn needed to create the Deployer
# class (ie: it uses ArgoWorkflowsDeployer to create the Deployer class).


class DeployerImpl(object):
    """
    Base class for deployer implementations. Each implementation should define a TYPE
    class variable that matches the name of the CLI group.

    Parameters
    ----------
    flow_file : str
        Path to the flow file to deploy, relative to current directory.
    show_output : bool, default True
        Show the 'stdout' and 'stderr' to the console by default.
    profile : Optional[str], default None
        Metaflow profile to use for the deployment. If not specified, the default
        profile is used.
    env : Optional[Dict], default None
        Additional environment variables to set for the deployment.
    cwd : Optional[str], default None
        The directory to run the subprocess in; if not specified, the current
        directory is used.
    file_read_timeout : int, default 3600
        The timeout until which we try to read the deployer attribute file (in seconds).
    **kwargs : Any
        Additional arguments that you would pass to `python myflow.py` before
        the deployment command.
    """

    TYPE: ClassVar[Optional[str]] = None

    def __init__(
        self,
        flow_file: str,
        show_output: bool = True,
        profile: Optional[str] = None,
        env: Optional[Dict] = None,
        cwd: Optional[str] = None,
        file_read_timeout: int = 3600,
        **kwargs
    ):
        if self.TYPE is None:
            raise ValueError(
                "DeployerImpl doesn't have a 'TYPE' to target. Please use a sub-class "
                "of DeployerImpl."
            )

        from metaflow.parameters import flow_context

        # Reload the CLI with an "empty" flow -- this will remove any configuration
        # and parameter options. They are re-added in from_cli (called below).
        with flow_context(None):
            [
                importlib.reload(sys.modules[module])
                for module in self.to_reload
                if module in sys.modules
            ]

        from metaflow.cli import start
        from metaflow.runner.click_api import MetaflowAPI

        # Convert flow_file to absolute path if it's relative
        if not os.path.isabs(flow_file):
            self.flow_file = os.path.abspath(flow_file)
        else:
            self.flow_file = flow_file
        self.show_output = show_output
        self.profile = profile
        self.env = env
        self.cwd = cwd or os.getcwd()
        self.file_read_timeout = file_read_timeout

        self.env_vars = os.environ.copy()
        self.env_vars.update(self.env or {})
        if self.profile:
            self.env_vars["METAFLOW_PROFILE"] = profile

        self.spm = SubprocessManager()
        self.top_level_kwargs = kwargs
        self.api = MetaflowAPI.from_cli(self.flow_file, start)

    @property
    def to_reload(self) -> List[str]:
        """
        List of modules to reload when the deployer is initialized.
        This is used to ensure that the CLI is in a clean state before
        deploying the flow.
        """
        return [
            "metaflow.cli",
            "metaflow.cli_components.run_cmds",
            "metaflow.cli_components.init_cmd",
        ]

    @property
    def deployer_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError

    @staticmethod
    def deployed_flow_type() -> Type["metaflow.runner.deployer.DeployedFlow"]:
        raise NotImplementedError

    def __enter__(self) -> "DeployerImpl":
        return self

    def create(self, **kwargs) -> "metaflow.runner.deployer.DeployedFlow":
        """
        Create a sub-class of a `DeployedFlow` depending on the deployer implementation.

        Parameters
        ----------
        **kwargs : Any
            Additional arguments to pass to `create` corresponding to the
            command line arguments of `create`

        Returns
        -------
        DeployedFlow
            DeployedFlow object representing the deployed flow.

        Raises
        ------
        Exception
            If there is an error during deployment.
        """
        # Sub-classes should implement this by simply calling _create and pass the
        # proper class as the DeployedFlow to return.
        raise NotImplementedError

    def _create(
        self, create_class: Type["metaflow.runner.deployer.DeployedFlow"], **kwargs
    ) -> "metaflow.runner.deployer.DeployedFlow":
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            # every subclass needs to have `self.deployer_kwargs`
            # TODO: Get rid of CLICK_API_PROCESS_CONFIG in the near future
            if CLICK_API_PROCESS_CONFIG:
                # We need to run this in the cwd because configs depend on files
                # that may be located in paths relative to the directory the user
                # wants to run in
                with with_dir(self.cwd):
                    command = get_lower_level_group(
                        self.api, self.top_level_kwargs, self.TYPE, self.deployer_kwargs
                    ).create(deployer_attribute_file=attribute_file_path, **kwargs)
            else:
                command = get_lower_level_group(
                    self.api, self.top_level_kwargs, self.TYPE, self.deployer_kwargs
                ).create(deployer_attribute_file=attribute_file_path, **kwargs)

            pid = self.spm.run_command(
                [sys.executable, *command],
                env=self.env_vars,
                cwd=self.cwd,
                show_output=self.show_output,
            )

            command_obj = self.spm.get(pid)
            content = handle_timeout(
                attribute_file_fd, command_obj, self.file_read_timeout
            )
            content = json.loads(content)
            self.name = content.get("name")
            self.flow_name = content.get("flow_name")
            self.metadata = content.get("metadata")
            # Additional info is used to pass additional deployer specific information.
            # It is used in non-OSS deployers (extensions).
            self.additional_info = content.get("additional_info", {})
            command_obj.sync_wait()
            if command_obj.process.returncode == 0:
                return create_class(deployer=self)

        raise RuntimeError("Error deploying %s to %s" % (self.flow_file, self.TYPE))

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Cleanup resources on exit.
        """
        self.cleanup()

    def cleanup(self):
        """
        Cleanup resources.
        """
        self.spm.cleanup()
